/***************************************************************************/
/* RSC IDENTIFIER: Source
 *
 * ABSTRACT
 *
 *
 * ERROR HANDLING
 *
 *    This component checks for input file errors and input parameter errors.
 *    If an invalid value is found, the error code is combined with the current
 *    error code using the bitwise or.  This combining allows multiple error
 *    codes to be returned. The possible error codes are:
 *
 *  SOURCE_NO_ERROR                  : No errors occurred in function
 *  SOURCE_INVALID_INDEX_ERROR       : Index out of valid range (less than one
 *                                      or more than Source_Count)
 *  SOURCE_INVALID_NAME_ERROR        : Source name not found
 *
 *
 * REUSE NOTES
 *
 *    Source is intended for reuse by any application that needs access to 
 *    accuracy sources.
 *
 *    
 * REFERENCES
 *
 *    Further information on Source can be found in the Reuse Manual.
 *
 *    Source originated from :  U.S. Army Topographic Engineering Center (USATEC)
 *                              Geospatial Information Division (GID)
 *                              7701 Telegraph Road
 *                              Alexandria, VA  22310-3864
 *
 * LICENSES
 *
 *    None apply to this component.
 *
 * RESTRICTIONS
 *
 *    Source has no restrictions.
 *
 * ENVIRONMENT
 *
 *    Source was tested and certified in the following environments:
 *
 *    1. Solaris 2.5 with GCC 2.8.1
 *    2. MS Windows 95 with MS Visual C++ 6
 *
 * MODIFICATIONS
 *
 *    Date              Description
 *    ----              -----------
 */


/***************************************************************************/
/*
 *                               INCLUDES
 */

#include <string.h>
#include <ctype.h>
#include "source.h"

/***************************************************************************/
/*
 *                               DEFINES
 */
#define SOURCE_NAME_LENGTH          50
#define NUMBER_SOURCES              19

/***************************************************************************/
/*
 *                          GLOBAL DECLARATIONS
 */

typedef struct Accuracy_Source_Value
{
  char Name[SOURCE_NAME_LENGTH];
  double CE;        /* circular error */
  double LE;        /* linear error */
  double SE;        /* spherical error */
} Accuracy_Sources; /* defines an accuracy source */


static const Accuracy_Sources Source_Table[NUMBER_SOURCES] = {{"Unknown", -1, -1, -1},
                                                              {"User defined (1 meter)", 1, 1, 1},
                                                              {"GPS PPS", 20, 20, 20},
                                                              {"GPS SPS", 20, 20, 20},
                                                              {"1:25,000 City Map", 50, 20, 50},
                                                              {"1:50,000 Topographic Line Map (TLM)", 50, 20, 50},
                                                              {"1:50,000 Combat Chart", 50, 20, 50},
                                                              {"1:100,000 Topographic Line Map (TLM)", 50, 20, 50},
                                                              {"1:250,000 Joint Operations Graphic (JOG)", 250, 100, 250},
                                                              {"1:500,000 Tactical Pilotage Chart (TPC)", 1000, 150, 1000},
                                                              {"1:1,000,000 Operational Navigation Chart (ONC)", 2000, 650, 2000},
                                                              {"Digital Terrain Elevation Data (DTED) Level 1", 30, 30, 30},
                                                              {"Digital Terrain Elevation Data (DTED) Level 2", 30, 30, 30},
                                                              {"Digital Feature Analysis Data (DFAD) Level 1", 130, -1, 130},
                                                              {"Digital Feature Analysis Data (DFAD) Level 2", 130, -1, 130},
                                                              {"Controlled Image Base (CIB) [>10m GSD]", 25, -1, 25},
                                                              {"Vector Smart Map (VMAP) Level 1", 250, 50, 250},
                                                              {"Vector Smart Map (VMAP) Level 2 (1:100,000)", 100, 20, 100},
                                                              {"Vector Smart Map (VMAP) Level 2 (1:50,000)", 50, 10, 50}};


/***************************************************************************/
/*
 *                              FUNCTIONS     
 */
void Source_Count (long *Count)

{ /* Begin Source_Count */
  /*
   *  The function Source_Count returns the number of source accuracies 
   *
   *  Count                : number of accuracy sources      (output)
   */

    *Count = NUMBER_SOURCES;

} /* End of Source_Count */


long Source_Index( const char *Name,
                   long *Index )

{ /* Begin Source_Index */
  /*
   *  The function Source_Index returns the index of the source accuracy with the 
   *  specified name.
   *
   *  Name    : The source name being searched for.                   (input)
   *  Index   : The index of the accuracy source with the             (output)
   *              specified code.
   */
  char temp_name[SOURCE_NAME_LENGTH];
  long error_code = SOURCE_NO_ERROR;
  long length;
  long i = 0;

  *Index = 0;
  length = strlen(Name);
  if (length > (SOURCE_NAME_LENGTH-1))
    error_code |= SOURCE_INVALID_NAME_ERROR;
  else
  {
    strcpy(temp_name,Name);

    /* Convert to upper case */
    for (i=0;i<length;i++)
      temp_name[i] = (char)toupper(temp_name[i]);

    /* Strip blank spaces */
/*      while (pos < length)
    {
      if (isspace(temp_name[pos]))
      {
        for (i=pos;i<=length;i++)
          temp_name[i] = temp_name[i+1];
        length -= 1;
      }
      else
        pos += 1;
    }*/
    /* Search for code */
    i = 0;
    while (i < NUMBER_SOURCES && strcmp(temp_name, Source_Table[i].Name))
    {
      i++;
    }
    if (i == NUMBER_SOURCES || strcmp(temp_name, Source_Table[i].Name))
      error_code |= SOURCE_INVALID_NAME_ERROR;
    else
      *Index = i+1;
  }
  return (error_code);
} /* End Source_Index */


long Source_Name (const long Index,
                  char *Name)
{ /* Begin Source_Name */
  /*
   *  The function Source_Name returns the name of the source accuracy referenced by
   *  index.
   *
   *  Index   : The index of a given datum in the datum table.        (input)
   *  Name    : The accuracy source Name referenced by Index.         (output)
   */

  long error_code = SOURCE_NO_ERROR;

  if ((Index > 0) && (Index <= NUMBER_SOURCES))
    strcpy(Name, Source_Table[Index-1].Name);
  else
    error_code |= SOURCE_INVALID_INDEX_ERROR;
  return (error_code);
} /* End Source_Name */


long Source_Accuracy (const long Index, 
					            double *CE,
					            double *LE,
					            double *SE)
{ /* Begin Source_Accuracy */
/*
 *   The function Source_Accuracy returns the accuracy values (CE, LE, SE)
 *   for the source referenced by index.
 *
 *    Index      : The index of a given datum in the datum table   (input)
 *    CE         : Circular error                                  (output)
 *    LE         : Linear error                                    (output)
 *    SE         : Spherical error                                 (output)
 */

  long error_code = SOURCE_NO_ERROR;

  if (Index > 0 && Index <= NUMBER_SOURCES)
  {
    *CE = Source_Table[Index-1].CE;
    *LE = Source_Table[Index-1].LE;
    *SE = Source_Table[Index-1].SE;
  }
  else
  {
    error_code |= SOURCE_INVALID_INDEX_ERROR;
  }
  return (error_code);
} /* End Source_Accuracy */

